# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from collections import OrderedDict
from typing import Optional

from duniterpy.key import SigningKey
from PyQt5.QtCore import QMutex, QSize, QTimer
from PyQt5.QtGui import QFont, QMovie
from PyQt5.QtWidgets import QApplication, QLineEdit, QWidget, QWizard
from substrateinterface import Keypair, KeypairType

from tikka.domains.application import Application
from tikka.domains.entities.constants import (
    AMOUNT_UNIT_KEY,
    DATA_PATH,
    DERIVATION_PATH_MEMBER,
    WALLETS_PASSWORD_LENGTH,
)
from tikka.interfaces.adapters.repository.accounts import AccountsRepositoryInterface
from tikka.libs.secret import generate_alphabetic, sanitize_mnemonic_string
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    DEBOUNCE_TIME,
    ICON_LOADER,
    SELECTED_UNIT_PREFERENCES_KEY,
)
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.resources.gui.windows.v1_account_import_wizard_rc import (
    Ui_importAccountV1Wizard,
)


class V1AccountImportWizardWindow(QWizard, Ui_importAccountV1Wizard):
    """
    V1AccountImportWizardWindow class
    """

    def __init__(
        self, application: Application, mutex: QMutex, parent: Optional[QWidget] = None
    ):
        """
        Init import V1 account wizard window

        :param application: Application instance
        :param mutex: QMutex instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext
        self.mutex = mutex

        unit_preference = self.application.preferences_repository.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            self.amount = self.application.amounts.get_amount(unit_preference)
        else:
            self.amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)

        # source account
        self.source_account = None
        self.source_v1_address = None
        self.source_keypair = None
        self.source_wallet_password = generate_alphabetic(WALLETS_PASSWORD_LENGTH)

        # destination account
        self.destination_root_account = None
        self.destination_mnemonic = None
        self.destination_account = None
        self.destination_keypair = None

        self.display_identity_status = {
            "Created": self._("Created"),
            "Validated": self._("Validated"),
            "ConfirmedByOwner": self._("Confirmed"),
            "Unknown": self._("Unknown"),
        }

        # set monospace font to address fields
        self.monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        self.monospace_font.setStyleHint(QFont.Monospace)
        self.sourceAddressValueImportLabel.setFont(self.monospace_font)
        self.destinationAddressValueImportLabel.setFont(self.monospace_font)

        # animated loading icon
        self.loader_movie = QMovie(ICON_LOADER)
        self.loader_movie.setScaledSize(QSize(16, 16))
        self.loader_movie.start()

        # debounce timers
        self.source_debounce_timer = QTimer()
        self.destination_debounce_timer = QTimer()

        ##############################
        # ASYNC METHODS
        ##############################
        # Create a QWorker object
        self.fetch_source_from_network_async_qworker = AsyncQWorker(
            self.fetch_source_from_network, self.mutex
        )
        self.fetch_destination_from_network_async_qworker = AsyncQWorker(
            self.fetch_destination_from_network, self.mutex
        )
        self.import_source_into_destination_on_network_async_qworker = AsyncQWorker(
            self.import_source_into_destination_on_network, self.mutex
        )
        self.wizard_page2_init()

    def wizard_page2_init(self):
        """
        Initialize page 2 form

        :return:
        """
        # fonts
        self.sourceV1AddressValueLabel.setFont(self.monospace_font)
        self.sourceAddressValueLabel.setFont(self.monospace_font)

        # page next button status handling
        self.wizardPage2.isComplete = self.wizard_page2_is_complete

        # events
        self.sourceSecretIDLineEdit.textChanged.connect(
            self._on_source_secret_id_line_edit_changed
        )
        self.sourcePasswordIDLineEdit.textChanged.connect(
            self._on_source_password_id_line_edit_changed
        )
        self.sourceShowButton.clicked.connect(self.on_source_show_button_clicked)

        # debounce timer
        self.source_debounce_timer.timeout.connect(self._source_generate_address)

        ##############################
        # ASYNC METHODS
        ##############################
        self.fetch_source_from_network_async_qworker.finished.connect(
            self._on_finished_fetch_source_from_network
        )

    def wizard_page2_is_complete(self) -> bool:
        """
        Function to overload V1AccountImportWizardWindow->Page2->IsComplete() method

        :return:
        """
        result = False

        if (
            self.source_account is not None
            and self.source_keypair is not None
            and self.source_account.balance is not None
        ):
            result = True

        return result

    def _on_source_secret_id_line_edit_changed(self):
        """
        Triggered when text is changed in the secret ID field

        :return:
        """
        if self.source_debounce_timer.isActive():
            self.source_debounce_timer.stop()
        self.source_debounce_timer.start(DEBOUNCE_TIME)

    def _on_source_password_id_line_edit_changed(self):
        """
        Triggered when text is changed in the password ID field

        :return:
        """
        if self.source_debounce_timer.isActive():
            self.source_debounce_timer.stop()
        self.source_debounce_timer.start(DEBOUNCE_TIME)

    def _source_generate_address(self):
        """
        Generate address from ID

        :return:
        """
        # stop debounce_timer to avoid infinite loop
        if self.source_debounce_timer.isActive():
            self.source_debounce_timer.stop()

        self.sourceV1AddressValueLabel.setText("")
        self.sourceAddressValueLabel.setText("")
        self.sourceErrorLabel.setText("")

        secret_id = self.sourceSecretIDLineEdit.text().strip()
        password_id = self.sourcePasswordIDLineEdit.text().strip()
        if secret_id == "" or password_id == "":
            return

        signing_key = SigningKey.from_credentials(secret_id, password_id)
        self.source_v1_address = signing_key.pubkey
        self.sourceV1AddressValueLabel.setText(self.source_v1_address)

        try:
            self.source_keypair = Keypair.create_from_seed(
                seed_hex=signing_key.seed.hex(),
                ss58_format=self.application.currencies.get_current().ss58_format,
                crypto_type=KeypairType.ED25519,
            )
        except Exception as exception:
            logging.exception(exception)
            self.sourceErrorLabel.setText(self._("Error generating account wallet!"))
            self.sourceErrorLabel.setStyleSheet("color: red;")
            return

        self.source_account = self.application.accounts.get_instance(
            self.source_keypair.ss58_address
        )
        self.sourceAddressValueLabel.setText(self.source_account.address)

        # fetch if v1 account exists on network
        self.fetch_source_from_network_async_qworker.start()

    def on_source_show_button_clicked(self):
        """
        Triggered when user click on source show button

        :return:
        """
        if self.sourceSecretIDLineEdit.echoMode() == QLineEdit.Password:
            self.sourceSecretIDLineEdit.setEchoMode(QLineEdit.Normal)
            self.sourcePasswordIDLineEdit.setEchoMode(QLineEdit.Normal)
            self.sourceShowButton.setText(self._("Hide"))
        else:
            self.sourceSecretIDLineEdit.setEchoMode(QLineEdit.Password)
            self.sourcePasswordIDLineEdit.setEchoMode(QLineEdit.Password)
            self.sourceShowButton.setText(self._("Show"))

    def fetch_source_from_network(self):
        """
        Fetch last account data from the network

        :return:
        """
        self.sourceErrorLabel.setMovie(self.loader_movie)

        try:
            balance = self.application.accounts.fetch_balance_from_network(
                self.source_account.address
            )
        except Exception:
            # keep balance from cache DB if network failure
            balance = self.source_account.balance

        if balance is None:
            return

        self.source_account.balance = balance
        self.source_account.identity_index = (
            self.application.identities.fetch_index_from_network(
                self.source_account.address
            )
        )
        if self.source_account.identity_index is not None:
            self.application.identities.fetch_from_network(
                self.source_account.identity_index
            )

    def _on_finished_fetch_source_from_network(self):
        """
        Triggered when async request fetch_from_network is finished

        :return:
        """
        self.sourceErrorLabel.setMovie(None)

        self.application.accounts.update(self.source_account)
        if self.source_account is not None:
            if self.source_account.balance is None:
                self.sourceErrorLabel.setText(self._("This account does not exists"))
                self.sourceErrorLabel.setStyleSheet("color: red;")
            else:
                self.sourceBalanceValueLabel.setText(
                    self.locale().toCurrencyString(
                        self.amount.value(self.source_account.balance),
                        self.amount.symbol(),
                    )
                )

                if self.source_account.identity_index is not None:
                    identity = self.application.identities.get(
                        self.source_account.identity_index
                    )
                    if identity.status not in self.display_identity_status:
                        self.sourceIdentityValueLabel.setText(
                            self.display_identity_status["Unknown"]
                        )
                    else:
                        self.sourceIdentityValueLabel.setText(
                            self.display_identity_status[identity.status]
                        )
                else:
                    self.sourceIdentityValueLabel.setText(self._("None"))

                self.sourceErrorLabel.setText(self._("Account is valid"))
                self.sourceErrorLabel.setStyleSheet("color: green;")
        self.wizardPage2.completeChanged.emit()

        if self.wizardPage2.isComplete():
            self.wizard_page3_init()

    def wizard_page3_init(self):
        """
        Initialize page 3 form

        :return:
        """
        # Mnemonic language selector translated
        mnemonic_language_selector = OrderedDict(
            [
                ("en", self._("English")),
                ("fr", self._("French")),
                ("zh-hans", self._("Chinese simplified")),
                ("zh-hant", self._("Chinese traditional")),
                ("it", self._("Italian")),
                ("ja", self._("Japanese")),
                ("ko", self._("Korean")),
                ("es", self._("Spanish")),
            ]
        )
        for language_code, language_name in mnemonic_language_selector.items():
            self.destinationMnemonicLanguageComboBox.addItem(
                language_name, userData=language_code
            )
        self.destinationMnemonicLanguageComboBox.setCurrentIndex(
            self.destinationMnemonicLanguageComboBox.findData(
                self.application.config.get("language")[:2]
            )
        )

        # fonts
        self.destinationRootNameOrAddressComboBox.setFont(self.monospace_font)
        self.destinationRootAddressValueLabel.setFont(self.monospace_font)
        self.destinationAddressValueLabel.setFont(self.monospace_font)

        # page next button status handling
        self.wizardPage3.isComplete = self.wizard_page3_is_complete

        # init root accounts combo box
        # only V2 root accounts...
        for account in self.application.accounts.get_list(
            filters={
                AccountsRepositoryInterface.COLUMN_ROOT: True,
                AccountsRepositoryInterface.COLUMN_CRYPTO_TYPE: int(
                    KeypairType.SR25519
                ),
            }
        ):
            # only accounts with password (user owns it)
            if self.application.passwords.exists(account.address):
                self.destinationRootNameOrAddressComboBox.addItem(
                    account.name if account.name is not None else account.address,
                    userData=account.address,
                )

        # events
        self.destinationRootNameOrAddressComboBox.currentIndexChanged.connect(
            self._generate_destination_address
        )
        self.destinationMnemonicLineEdit.textChanged.connect(
            self._on_destination_mnemonic_line_edit_changed
        )
        self.destinationShowButton.clicked.connect(
            self.on_destination_show_button_clicked
        )
        self.destinationMnemonicLanguageComboBox.currentIndexChanged.connect(
            self._on_destination_mnemonic_line_edit_changed
        )
        # debounce timer
        self.destination_debounce_timer.timeout.connect(
            self._generate_destination_address
        )

        ##############################
        # ASYNC METHODS
        ##############################
        self.fetch_destination_from_network_async_qworker.finished.connect(
            self._on_finished_fetch_destination_from_network
        )

        self._generate_destination_address()

    def wizard_page3_is_complete(self) -> bool:
        """
        Function to overload V1AccountImportWizardWindow->Page3->IsComplete() method

        :return:
        """
        result = False

        if (
            self.destination_account is not None
            and self.destination_keypair is not None
            and (
                self.source_account.identity_index is None
                or (
                    self.source_account.identity_index is not None
                    and self.destination_account.identity_index is None
                )
            )
        ):
            result = True

        return result

    def _on_destination_mnemonic_line_edit_changed(self):
        """
        Triggered when text is changed in the mnemonic field

        :return:
        """
        if self.destination_debounce_timer.isActive():
            self.destination_debounce_timer.stop()
        self.destination_debounce_timer.start(DEBOUNCE_TIME)

    def on_destination_show_button_clicked(self):
        """
        Triggered when user click on destination show button

        :return:
        """
        if self.destinationMnemonicLineEdit.echoMode() == QLineEdit.Password:
            self.destinationMnemonicLineEdit.setEchoMode(QLineEdit.Normal)
            self.storedPasswordLineEdit.setEchoMode(QLineEdit.Normal)
            self.destinationShowButton.setText(self._("Hide"))
        else:
            self.destinationMnemonicLineEdit.setEchoMode(QLineEdit.Password)
            self.storedPasswordLineEdit.setEchoMode(QLineEdit.Password)
            self.destinationShowButton.setText(self._("Show"))

    def _generate_destination_address(self):
        """
        Generate destination address

        :return:
        """
        # stop debounce_timer to avoid infinite loop
        if self.destination_debounce_timer.isActive():
            self.destination_debounce_timer.stop()

        self.destination_root_account = self.application.accounts.get_by_address(
            self.destinationRootNameOrAddressComboBox.currentData()
        )

        # select derivation: //0 if source account is member, first even number if not
        destination_derivation = None
        if (
            self.source_account is not None
            and self.source_account.identity_index is not None
        ):
            destination_derivation = DERIVATION_PATH_MEMBER

        if self.destination_root_account is not None and destination_derivation is None:
            available_derivation_list = (
                self.application.accounts.get_available_derivation_list(
                    self.destination_root_account
                )
            )
            if available_derivation_list[0] == DERIVATION_PATH_MEMBER:
                destination_derivation = available_derivation_list[1]
            else:
                destination_derivation = available_derivation_list[0]

        self.destinationDerivationValueLabel.setText(destination_derivation)

        # if account name displayed in combobox...
        if (
            self.destinationRootNameOrAddressComboBox.currentText()
            != self.destinationRootNameOrAddressComboBox.currentData()
        ):
            self.destinationRootAddressValueLabel.setText(
                self.destinationRootNameOrAddressComboBox.currentData()
            )
        else:
            self.destinationRootAddressValueLabel.setText("")

        if self.destinationMnemonicLineEdit.text().strip() == "":
            return

        self.destinationErrorLabel.setText("")
        language_code = self.destinationMnemonicLanguageComboBox.currentData()
        self.destination_mnemonic = sanitize_mnemonic_string(
            self.destinationMnemonicLineEdit.text()
        )
        suri = self.destination_mnemonic + destination_derivation
        if not Keypair.validate_mnemonic(self.destination_mnemonic, language_code):
            self.destinationErrorLabel.setText(
                self._("Mnemonic or language not valid!")
            )
            self.destinationErrorLabel.setStyleSheet("color: red;")
            return

        try:
            root_keypair = Keypair.create_from_mnemonic(
                mnemonic=self.destination_mnemonic,
                ss58_format=self.application.currencies.get_current().ss58_format,
                crypto_type=KeypairType.SR25519,
                language_code=language_code,
            )
        except Exception as exception:
            logging.exception(exception)
            self.destinationErrorLabel.setText(
                self._("Mnemonic or language not valid!")
            )
            self.destinationErrorLabel.setStyleSheet("color: red;")
            return

        # if mnemonic address is not account address...
        if root_keypair.ss58_address != self.destination_root_account.address:
            self.destinationErrorLabel.setText(
                self._("Mnemonic address is not the root account address!")
            )
            self.destinationErrorLabel.setStyleSheet("color: red;")
            return

        stored_password = self.application.passwords.get_clear_password(root_keypair)
        self.storedPasswordLineEdit.setText(stored_password)

        try:
            self.destination_keypair = Keypair.create_from_uri(
                suri=suri,
                ss58_format=self.application.currencies.get_current().ss58_format,
                crypto_type=KeypairType.SR25519,
                language_code=language_code,
            )
        except Exception as exception:
            logging.exception(exception)
            self.destinationErrorLabel.setText(
                self._("Mnemonic or language not valid!")
            )
            self.destinationErrorLabel.setStyleSheet("color: red;")
            return

        if self.destination_keypair is None:
            return

        self.destination_account = self.application.accounts.get_by_address(
            self.destination_keypair.ss58_address
        )
        if self.destination_account is None:
            self.destination_account = self.application.accounts.get_instance(
                self.destination_keypair.ss58_address,
                self.destinationNameLineEdit.text().strip(),
            )
            self.destination_account.crypto_type = KeypairType.SR25519
            self.destination_account.root = self.destination_root_account.address
            self.destination_account.path = destination_derivation
        self.destinationAddressValueLabel.setText(self.destination_account.address)

        # fetch if v2 account exists on network
        self.fetch_destination_from_network_async_qworker.start()

    def fetch_destination_from_network(self):
        """
        Fetch destination account data from the network

        :return:
        """
        self.destinationErrorLabel.setMovie(self.loader_movie)

        try:
            balance = self.application.accounts.fetch_balance_from_network(
                self.destination_account.address
            )
        except Exception:
            # keep balance from cache DB if network failure
            balance = self.destination_account.balance

        if balance is None:
            return

        self.destination_account.balance = balance
        self.destination_account.identity_index = (
            self.application.identities.fetch_index_from_network(
                self.destination_account.address
            )
        )
        if self.destination_account.identity_index is not None:
            self.application.identities.fetch_from_network(
                self.destination_account.identity_index
            )

    def _on_finished_fetch_destination_from_network(self):
        """
        Triggered when async request fetch_from_network is finished

        :return:
        """
        self.destinationErrorLabel.setMovie(None)

        if self.destination_account.balance is not None:
            self.destinationBalanceValueLabel.setText(
                self.locale().toCurrencyString(
                    self.amount.value(self.destination_account.balance),
                    self.amount.symbol(),
                )
            )

            if self.destination_account.identity_index is not None:
                self.destinationErrorLabel.setText(
                    self._("Account has already an identity!")
                )
                self.destinationErrorLabel.setStyleSheet("color: red;")
            else:
                self.destinationErrorLabel.setText(self._("Account is valid"))
                self.destinationErrorLabel.setStyleSheet("color: green;")

        self.wizardPage3.completeChanged.emit()

        if self.wizardPage3.isComplete():
            self.wizard_page4_init()

    def wizard_page4_init(self):
        """
        Initialize page 4 form

        :return:
        """
        # fonts
        self.sourceAddressValueImportLabel.setFont(self.monospace_font)
        self.destinationAddressValueImportLabel.setFont(self.monospace_font)

        self.sourceAddressValueImportLabel.setText(self.source_v1_address)
        if self.source_account is not None:
            if self.source_account.balance is not None:
                self.sourceBalanceValueImportLabel.setText(
                    self.locale().toCurrencyString(
                        self.amount.value(self.source_account.balance),
                        self.amount.symbol(),
                    )
                )
        if self.source_account.identity_index is not None:
            identity = self.application.identities.get(
                self.source_account.identity_index
            )
            if identity.status not in self.display_identity_status:
                self.sourceIdentityValueImportLabel.setText(
                    self.display_identity_status["Unknown"]
                )
            else:
                self.sourceIdentityValueImportLabel.setText(
                    self.display_identity_status[identity.status]
                )
        else:
            self.sourceIdentityValueImportLabel.setText(self._("None"))

        self.destinationAddressValueImportLabel.setText(
            self.destination_account.address
        )
        if self.destination_account.balance is not None:
            self.destinationBalanceValueImportLabel.setText(
                self.locale().toCurrencyString(
                    self.amount.value(self.destination_account.balance),
                    self.amount.symbol(),
                )
            )
        self.destinationIdentityValueImportLabel.setText(self._("None"))

        # page next button status handling
        self.wizardPage4.isComplete = self.wizard_page4_is_complete

        # events
        self.importButton.clicked.connect(self.on_import_button_clicked)

        ##############################
        # ASYNC METHODS
        ##############################
        self.import_source_into_destination_on_network_async_qworker.finished.connect(
            self._on_finished_import_source_into_destination_on_network
        )

    def wizard_page4_is_complete(self) -> bool:
        """
        Function to overload V1AccountImportWizardWindow->Page4->IsComplete() method

        :return:
        """
        result = False

        if (
            self.source_account is not None
            and (
                self.source_account.balance == 0 or self.source_account.balance is None
            )
            and self.source_account.identity_index is None
        ):
            result = True

        return result

    def on_import_button_clicked(self):
        """
        Triggered when user click on import button

        :return:
        """
        self.importButton.setDisabled(True)
        self.importErrorLabel.setMovie(self.loader_movie)

        self.import_source_into_destination_on_network_async_qworker.start()

    def import_source_into_destination_on_network(self):
        """
        Send changeOwnerKey for identity if any and tranfer money from V1 source account to V2 destination account

        :return:
        """
        if self.source_keypair is None:
            return

        if self.source_account.identity_index is not None:
            result = self.application.identities.change_owner_key(
                self.source_keypair, self.destination_keypair
            )
            if not result:
                # do not transfer all the money if identity transfer failed,
                # because identity transfer requires fees from sender account
                return

        if self.source_account.balance is not None and self.source_account.balance > 0:
            self.application.transfers.send_all(
                self.source_keypair, self.destination_account.address, False
            )

        try:
            balance = self.application.accounts.fetch_balance_from_network(
                self.source_account.address
            )
        except Exception:
            # keep balance from cache DB if network failure
            balance = self.source_account.balance

        self.source_account.balance = balance

        self.source_account.identity_index = (
            self.application.identities.fetch_index_from_network(
                self.source_account.address
            )
        )

        try:
            balance = self.application.accounts.fetch_balance_from_network(
                self.destination_account.address
            )
        except Exception:
            # keep balance from cache DB if network failure
            balance = self.destination_account.balance

        self.destination_account.balance = balance

        self.destination_account.identity_index = (
            self.application.identities.fetch_index_from_network(
                self.destination_account.address
            )
        )

    def _on_finished_import_source_into_destination_on_network(self):
        """
        Triggered when async request import_source_into_destination_on_network is finished

        :return:
        """
        self.importErrorLabel.setMovie(None)

        if self.source_account is not None:
            if self.source_account.balance is not None:
                self.sourceBalanceValueImportLabel.setText(
                    self.locale().toCurrencyString(
                        self.amount.value(self.source_account.balance),
                        self.amount.symbol(),
                    )
                )
            else:
                # empty account deleted from blockchain
                self.sourceBalanceValueImportLabel.setText("")

        if self.source_account.identity_index is not None:
            identity = self.application.identities.get(
                self.source_account.identity_index
            )
            if identity.status not in self.display_identity_status:
                self.sourceIdentityValueImportLabel.setText(
                    self.display_identity_status["Unknown"]
                )
            else:
                self.sourceIdentityValueImportLabel.setText(
                    self.display_identity_status[identity.status]
                )
        else:
            self.sourceIdentityValueImportLabel.setText(self._("None"))

        if self.destination_account.balance is not None:
            self.destinationBalanceValueImportLabel.setText(
                self.locale().toCurrencyString(
                    self.amount.value(self.destination_account.balance),
                    self.amount.symbol(),
                )
            )
        else:
            self.destinationBalanceValueImportLabel.setText("")

        if self.destination_account.identity_index is not None:
            identity = self.application.identities.get(
                self.destination_account.identity_index
            )
            if identity.status not in self.display_identity_status:
                self.destinationIdentityValueImportLabel.setText(
                    self.display_identity_status["Unknown"]
                )
            else:
                self.destinationIdentityValueImportLabel.setText(
                    self.display_identity_status[identity.status]
                )
        else:
            self.destinationIdentityValueImportLabel.setText(self._("None"))

        self.wizardPage4.completeChanged.emit()

        if self.wizardPage4.isComplete():
            self.importButton.setDisabled(True)
            self.importErrorLabel.setStyleSheet("color: green;")
            self.importErrorLabel.setText(self._("Account imported successfully!"))
            if (
                self.application.accounts.get_by_address(
                    self.destination_account.address
                )
                is None
            ):
                self.application.accounts.add(self.destination_account)
            else:
                self.application.accounts.update(self.destination_account)
        else:
            self.importButton.setDisabled(False)
            self.importErrorLabel.setStyleSheet("color: red;")
            if (
                self.source_account.balance is not None
                and self.source_account.balance > 0
            ):
                self.importErrorLabel.setText(self._("Error importing money!"))
            elif self.destination_account.identity_index is None:
                self.importErrorLabel.setText(self._("Error importing identity!"))


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    V1AccountImportWizardWindow(application_, QMutex()).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

from substrateinterface import Keypair

from tikka.domains.entities.node import Node
from tikka.domains.nodes import Nodes
from tikka.interfaces.adapters.network.authorities import NetworkAuthoritiesInterface


class Authorities:

    """
    Authorities domain class
    """

    def __init__(
        self,
        nodes: Nodes,
        network: NetworkAuthoritiesInterface,
    ):
        """
        Init Authorities domain

        :param nodes: Nodes domain instance
        :param network: NetworkAuthoritiesInterface instance
        """
        self.nodes = nodes
        self.network = network

    def rotate_keys(self, node: Node) -> Optional[str]:
        """
        Change node session keys and return them

        :param node: Node instance
        :return:
        """
        session_keys = self.network.rotate_keys()
        if session_keys is not None:
            node.session_keys = session_keys
            self.nodes.update(node)
        return session_keys

    def has_session_keys(self, session_keys: str) -> Optional[bool]:
        """
        Return True if the current node keystore store private session keys corresponding to public session_keys

        :param session_keys: Session public keys (hex string "0x123XYZ")
        :return:
        """
        return self.network.has_session_keys(session_keys)

    def publish_session_keys(self, keypair: Keypair, session_keys: str) -> bool:
        """
        Set/Change in blockchain the session public keys for the Keypair account

        :param keypair: Owner Keypair
        :param session_keys: Session public keys (hex string "0x123XYZ")
        :return:
        """
        return self.network.publish_session_keys(keypair, session_keys)

    def go_online(self, keypair: Keypair) -> bool:
        """
        Start writing blocks with smith account from keypair

        :param keypair: Smith account Keypair
        :return:
        """
        return self.network.go_online(keypair)

    def go_offline(self, keypair: Keypair) -> bool:
        """
        Stop writing blocks with smith account from keypair

        :param keypair: Smith account Keypair
        :return:
        """
        return self.network.go_offline(keypair)

    def fetch_is_online_from_network(self, identity_index: int) -> Optional[bool]:
        """
        Return True if identity_index is in online authorities

        :param identity_index: Identity index
        :return:
        """
        return self.network.is_online(identity_index)

    def fetch_is_incoming_from_network(self, identity_index: int) -> Optional[bool]:
        """
        Return True if identity_index is in incoming authorities

        :param identity_index: Identity index
        :return:
        """
        return self.network.is_incoming(identity_index)

    def fetch_is_outgoing_from_network(self, identity_index: int) -> Optional[bool]:
        """
        Return True if identity_index is in outgoing authorities

        :param identity_index: Identity index
        :return:
        """
        return self.network.is_outgoing(identity_index)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from typing import Optional

from tikka.interfaces.adapters.network.accounts import NetworkAccountsInterface


class NetworkAccounts(NetworkAccountsInterface):
    """
    NetworkAccounts class
    """

    def get_balance(self, address: str) -> Optional[int]:
        # __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
        #     NetworkAccountsInterface.get_balance.__doc__
        # )
        if not self.connections.is_connected():
            raise Exception("""Server not connected""")
        if self.connections.rpc.client is None:
            raise Exception("""Server not connected""")

        # system.account: FrameSystemAccountInfo
        # {
        #   nonce: 1
        #   consumers: 0
        #   providers: 1
        #   sufficients: 0
        #   data: {
        #     randomId: 0x18a4d...
        #     free: 9,799
        #     reserved: 0
        #     feeFrozen: 0
        #   }
        # }
        try:
            result = self.connections.rpc.client.query("System", "Account", [address])
        except Exception as exception:
            logging.exception(exception)
            raise exception

        if result.meta_info["result_found"] is False:
            return None

        return result.value["data"]["free"]

    get_balance.__doc__ = (  # pylint: disable=redefined-builtin, unused-variable
        NetworkAccountsInterface.get_balance.__doc__
    )



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/windows/main_window.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(843, 605)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())
        MainWindow.setSizePolicy(sizePolicy)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.centralwidget.sizePolicy().hasHeightForWidth())
        self.centralwidget.setSizePolicy(sizePolicy)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.tabWidget = QtWidgets.QTabWidget(self.centralwidget)
        self.tabWidget.setTabsClosable(True)
        self.tabWidget.setMovable(True)
        self.tabWidget.setObjectName("tabWidget")
        self.verticalLayout.addWidget(self.tabWidget)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 843, 25))
        self.menubar.setObjectName("menubar")
        self.accountMenu = QtWidgets.QMenu(self.menubar)
        self.accountMenu.setObjectName("accountMenu")
        self.helpMenu = QtWidgets.QMenu(self.menubar)
        self.helpMenu.setObjectName("helpMenu")
        self.menuNetwork = QtWidgets.QMenu(self.menubar)
        self.menuNetwork.setObjectName("menuNetwork")
        self.menuV1_accounts = QtWidgets.QMenu(self.menubar)
        self.menuV1_accounts.setObjectName("menuV1_accounts")
        self.menuAdvanced = QtWidgets.QMenu(self.menubar)
        self.menuAdvanced.setObjectName("menuAdvanced")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionAccount_tree = QtWidgets.QAction(MainWindow)
        self.actionAccount_tree.setObjectName("actionAccount_tree")
        self.actionAdd_an_address = QtWidgets.QAction(MainWindow)
        self.actionAdd_an_address.setObjectName("actionAdd_an_address")
        self.actionCreate_account = QtWidgets.QAction(MainWindow)
        self.actionCreate_account.setObjectName("actionCreate_account")
        self.actionV1Import_file = QtWidgets.QAction(MainWindow)
        self.actionV1Import_file.setObjectName("actionV1Import_file")
        self.actionQuit = QtWidgets.QAction(MainWindow)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/quit_icon"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionQuit.setIcon(icon)
        self.actionQuit.setObjectName("actionQuit")
        self.actionG1_licence = QtWidgets.QAction(MainWindow)
        self.actionG1_licence.setObjectName("actionG1_licence")
        self.actionConfiguration = QtWidgets.QAction(MainWindow)
        self.actionConfiguration.setObjectName("actionConfiguration")
        self.actionAbout = QtWidgets.QAction(MainWindow)
        self.actionAbout.setObjectName("actionAbout")
        self.actionConnection = QtWidgets.QAction(MainWindow)
        self.actionConnection.setObjectName("actionConnection")
        self.actionImport_account = QtWidgets.QAction(MainWindow)
        self.actionImport_account.setObjectName("actionImport_account")
        self.actionV1Import_account = QtWidgets.QAction(MainWindow)
        self.actionV1Import_account.setObjectName("actionV1Import_account")
        self.actionCurrency = QtWidgets.QAction(MainWindow)
        self.actionCurrency.setObjectName("actionCurrency")
        self.actionAdd_node = QtWidgets.QAction(MainWindow)
        self.actionAdd_node.setObjectName("actionAdd_node")
        self.actionNodes = QtWidgets.QAction(MainWindow)
        self.actionNodes.setObjectName("actionNodes")
        self.actionAccount_table = QtWidgets.QAction(MainWindow)
        self.actionAccount_table.setObjectName("actionAccount_table")
        self.actionWelcome = QtWidgets.QAction(MainWindow)
        self.actionWelcome.setObjectName("actionWelcome")
        self.actionTransfer = QtWidgets.QAction(MainWindow)
        self.actionTransfer.setObjectName("actionTransfer")
        self.actionImport_in_V2_wizard = QtWidgets.QAction(MainWindow)
        self.actionImport_in_V2_wizard.setObjectName("actionImport_in_V2_wizard")
        self.actionScan_a_QRCode = QtWidgets.QAction(MainWindow)
        self.actionScan_a_QRCode.setObjectName("actionScan_a_QRCode")
        self.actionSmith = QtWidgets.QAction(MainWindow)
        self.actionSmith.setObjectName("actionSmith")
        self.accountMenu.addAction(self.actionTransfer)
        self.accountMenu.addAction(self.actionAccount_tree)
        self.accountMenu.addAction(self.actionAccount_table)
        self.accountMenu.addSeparator()
        self.accountMenu.addAction(self.actionScan_a_QRCode)
        self.accountMenu.addAction(self.actionAdd_an_address)
        self.accountMenu.addAction(self.actionImport_account)
        self.accountMenu.addAction(self.actionCreate_account)
        self.accountMenu.addSeparator()
        self.accountMenu.addAction(self.actionQuit)
        self.helpMenu.addAction(self.actionWelcome)
        self.helpMenu.addAction(self.actionCurrency)
        self.helpMenu.addAction(self.actionG1_licence)
        self.helpMenu.addAction(self.actionConfiguration)
        self.helpMenu.addSeparator()
        self.helpMenu.addAction(self.actionAbout)
        self.menuNetwork.addAction(self.actionConnection)
        self.menuNetwork.addAction(self.actionNodes)
        self.menuNetwork.addAction(self.actionAdd_node)
        self.menuV1_accounts.addAction(self.actionImport_in_V2_wizard)
        self.menuV1_accounts.addAction(self.actionV1Import_account)
        self.menuV1_accounts.addAction(self.actionV1Import_file)
        self.menuAdvanced.addAction(self.actionSmith)
        self.menubar.addAction(self.accountMenu.menuAction())
        self.menubar.addAction(self.menuV1_accounts.menuAction())
        self.menubar.addAction(self.menuNetwork.menuAction())
        self.menubar.addAction(self.menuAdvanced.menuAction())
        self.menubar.addAction(self.helpMenu.menuAction())

        self.retranslateUi(MainWindow)
        self.tabWidget.setCurrentIndex(-1)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        #  _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_("MainWindow"))
        self.accountMenu.setTitle(_("Acco&unts"))
        self.helpMenu.setTitle(_("&Help"))
        self.menuNetwork.setTitle(_("Net&work"))
        self.menuV1_accounts.setTitle(_("&V1 accounts"))
        self.menuAdvanced.setTitle(_("Advanc&ed"))
        self.actionAccount_tree.setText(_("Account tr&ee"))
        self.actionAccount_tree.setIconText(_("Account tree"))
        self.actionAdd_an_address.setText(_("A&dd an address"))
        self.actionAdd_an_address.setIconText(_("Add an address"))
        self.actionCreate_account.setText(_("&Create an account"))
        self.actionV1Import_file.setText(_("Import &file"))
        self.actionV1Import_file.setIconText(_("Import file"))
        self.actionQuit.setText(_("&Quit"))
        self.actionG1_licence.setText(_("Ğ1 &licence"))
        self.actionConfiguration.setText(_("&Configuration"))
        self.actionAbout.setText(_("&About"))
        self.actionConnection.setText(_("&Connection"))
        self.actionConnection.setIconText(_("Connection"))
        self.actionImport_account.setText(_("&Import account"))
        self.actionImport_account.setIconText(_("Import account"))
        self.actionV1Import_account.setText(_("Import &account"))
        self.actionV1Import_account.setIconText(_("Import account"))
        self.actionCurrency.setText(_("C&urrency"))
        self.actionAdd_node.setText(_("&Add server"))
        self.actionAdd_node.setIconText(_("Add server"))
        self.actionNodes.setText(_("&Servers"))
        self.actionAccount_table.setText(_("Account ta&ble"))
        self.actionWelcome.setText(_("&Welcome"))
        self.actionTransfer.setText(_("&Transfer"))
        self.actionImport_in_V2_wizard.setText(_("&Import V1 in V2 wizard"))
        self.actionImport_in_V2_wizard.setIconText(_("Import V1 in V2 wizard"))
        self.actionScan_a_QRCode.setText(_("&Scan a QRCode"))
        self.actionSmith.setText(_("&Smith"))
import tikka.slots.pyqt.resources.icons.index_rc



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/widgets/connection.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ConnectionWidget(object):
    def setupUi(self, ConnectionWidget):
        ConnectionWidget.setObjectName("ConnectionWidget")
        ConnectionWidget.resize(724, 139)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(ConnectionWidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.urlsComboBox = QtWidgets.QComboBox(ConnectionWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.urlsComboBox.sizePolicy().hasHeightForWidth())
        self.urlsComboBox.setSizePolicy(sizePolicy)
        self.urlsComboBox.setToolTip("")
        self.urlsComboBox.setEditable(False)
        self.urlsComboBox.setMinimumContentsLength(0)
        self.urlsComboBox.setObjectName("urlsComboBox")
        self.gridLayout.addWidget(self.urlsComboBox, 0, 1, 1, 1)
        self.connectButton = QtWidgets.QPushButton(ConnectionWidget)
        self.connectButton.setObjectName("connectButton")
        self.gridLayout.addWidget(self.connectButton, 0, 2, 1, 1)
        self.connectionStatusLabel = QtWidgets.QLabel(ConnectionWidget)
        self.connectionStatusLabel.setMaximumSize(QtCore.QSize(16, 16))
        self.connectionStatusLabel.setText("")
        self.connectionStatusLabel.setPixmap(QtGui.QPixmap(":/icons/disconnected"))
        self.connectionStatusLabel.setScaledContents(True)
        self.connectionStatusLabel.setObjectName("connectionStatusLabel")
        self.gridLayout.addWidget(self.connectionStatusLabel, 0, 3, 1, 1)
        self.refreshNodeButton = QtWidgets.QPushButton(ConnectionWidget)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/refresh_icon"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.refreshNodeButton.setIcon(icon)
        self.refreshNodeButton.setObjectName("refreshNodeButton")
        self.gridLayout.addWidget(self.refreshNodeButton, 0, 4, 1, 1)
        self.versionValueLabel = QtWidgets.QLabel(ConnectionWidget)
        self.versionValueLabel.setText("")
        self.versionValueLabel.setObjectName("versionValueLabel")
        self.gridLayout.addWidget(self.versionValueLabel, 2, 1, 1, 1)
        self.peerIDLabel = QtWidgets.QLabel(ConnectionWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.peerIDLabel.setFont(font)
        self.peerIDLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.peerIDLabel.setObjectName("peerIDLabel")
        self.gridLayout.addWidget(self.peerIDLabel, 3, 0, 1, 1)
        self.serverLabel = QtWidgets.QLabel(ConnectionWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.serverLabel.setFont(font)
        self.serverLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.serverLabel.setObjectName("serverLabel")
        self.gridLayout.addWidget(self.serverLabel, 0, 0, 1, 1)
        self.peerIDValueLabel = QtWidgets.QLabel(ConnectionWidget)
        self.peerIDValueLabel.setText("")
        self.peerIDValueLabel.setObjectName("peerIDValueLabel")
        self.gridLayout.addWidget(self.peerIDValueLabel, 3, 1, 1, 1)
        self.softwareValueLabel = QtWidgets.QLabel(ConnectionWidget)
        self.softwareValueLabel.setText("")
        self.softwareValueLabel.setObjectName("softwareValueLabel")
        self.gridLayout.addWidget(self.softwareValueLabel, 1, 1, 1, 1)
        self.softwareLabel = QtWidgets.QLabel(ConnectionWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.softwareLabel.setFont(font)
        self.softwareLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.softwareLabel.setObjectName("softwareLabel")
        self.gridLayout.addWidget(self.softwareLabel, 1, 0, 1, 1)
        self.versionLabel = QtWidgets.QLabel(ConnectionWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.versionLabel.setFont(font)
        self.versionLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.versionLabel.setObjectName("versionLabel")
        self.gridLayout.addWidget(self.versionLabel, 2, 0, 1, 1)
        self.blockLabel = QtWidgets.QLabel(ConnectionWidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.blockLabel.setFont(font)
        self.blockLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.blockLabel.setObjectName("blockLabel")
        self.gridLayout.addWidget(self.blockLabel, 4, 0, 1, 1)
        self.blockValueLabel = QtWidgets.QLabel(ConnectionWidget)
        self.blockValueLabel.setText("")
        self.blockValueLabel.setObjectName("blockValueLabel")
        self.gridLayout.addWidget(self.blockValueLabel, 4, 1, 1, 1)
        self.verticalLayout_2.addLayout(self.gridLayout)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)

        self.retranslateUi(ConnectionWidget)
        QtCore.QMetaObject.connectSlotsByName(ConnectionWidget)

    def retranslateUi(self, ConnectionWidget):
        #  _translate = QtCore.QCoreApplication.translate
        ConnectionWidget.setWindowTitle(_("Connection"))
        self.connectButton.setToolTip(_("Connection to entry point"))
        self.connectButton.setText(_("Connect"))
        self.refreshNodeButton.setToolTip(_("Refresh nodes from network"))
        self.refreshNodeButton.setText(_("Refresh"))
        self.peerIDLabel.setText(_("Peer ID"))
        self.serverLabel.setText(_("Server"))
        self.softwareLabel.setText(_("Software"))
        self.versionLabel.setText(_("Version"))
        self.blockLabel.setText(_("Block"))
import tikka.slots.pyqt.resources.icons.index_rc



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import pickle
import re
from typing import Any, Iterable, List, Optional, Tuple, Type, Union
from uuid import UUID

from PyQt5.QtCore import QAbstractItemModel, QMimeData, QModelIndex, Qt, QVariant
from PyQt5.QtGui import QFont, QIcon, QPixmap

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.category import Category
from tikka.domains.entities.identity import STATUS_VALIDATED
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    ICON_ACCOUNT_NO_WALLET,
    ICON_ACCOUNT_WALLET_LOCKED,
    ICON_ACCOUNT_WALLET_UNLOCKED,
    ICON_IDENTITY,
)

MemberStatusRole = Qt.UserRole + 1000


class TreeItem:
    def __init__(
        self, element: Optional[Union[Category, Account]], parent: Optional[TreeItem]
    ) -> None:
        """
        Init TreeItem instance with element and parent

        :param element: Category or Account instance
        :param parent: TreeItem instance
        """
        self.element = element
        self._children: List[TreeItem] = []
        self._parent = parent
        self._row = 0

    def child_count(self):
        """
        Return number of children

        :return:
        """
        return len(self._children)

    def child(self, row) -> Optional[TreeItem]:
        """
        Return child at row

        :param row: Row number
        :return:
        """
        if 0 <= row < self.child_count():
            return self._children[row]
        return None

    def parent(self) -> Optional[TreeItem]:
        """
        Return parent

        :return:
        """
        return self._parent

    def row(self):
        """
        Return row number in parent's children

        :return:
        """
        return self._row

    def insert_child(self, position: int, child: TreeItem) -> bool:
        """
        Insert child at position row

        :param position: Row to insert child
        :param child: TreeItem instance
        :return:
        """
        self._children.insert(position, child)
        return True

    def add_child(self, child: TreeItem) -> None:
        """
        Add a child to item

        :param child: TreeItem instance
        :return:
        """
        child.set_parent(self)
        child.set_row(len(self._children))
        self._children.append(child)

    def set_parent(self, parent: TreeItem) -> None:
        """
        Set parent item

        :param parent: TreeItem instance
        :return:
        """
        self._parent = parent

    def set_row(self, row: int) -> None:
        """
        Set row value

        :param row: Row value
        :return:
        """
        self._row = row


class AccountTreeModel(QAbstractItemModel):
    """
    AccountTreeModel class that drives the population of tree display
    """

    def __init__(self, application: Application):
        super().__init__()

        self.application = application
        self._ = self.application.translator.gettext

        # drag and drop mime-type
        self.mime_type = "application/vnd.text.list"

        # Number of column displayed, see self.data()
        self._column_count = 2

        self.root_item = TreeItem(None, None)
        self.categories: List[Category] = []
        self.accounts: List[Account] = []
        self.init_data()

    def init_data(self):
        """
        Fill data from repository

        :return:
        """
        self.beginResetModel()
        self.root_item = TreeItem(None, None)
        self.categories = self.application.categories.list_all()
        self.accounts = self.application.accounts.get_list()
        self.recursive_append_children(None, self.root_item)
        self.endResetModel()

    def account_list_by_category_id(self, category_id: Optional[UUID]) -> List[Account]:
        """
        Return all accounts in category_id

        :param category_id: Category ID
        :return:
        """
        result = []
        for account in self.accounts:
            # do not list derived accounts
            if account.category_id is None and account.root is not None:
                continue
            if account.category_id == category_id:
                result.append(account)

        return result

    def category_children(self, category_id: Optional[UUID]) -> List[Category]:
        """
        Return categories with parent category_id

        :param category_id: Parent category ID
        :return:
        """
        return [
            category
            for category in self.categories
            if category.parent_id == category_id
        ]

    def get_derivation_accounts(self, address: str) -> List[Account]:
        """
        Return derived accounts from root address

        :param address: Root account address
        :return:
        """
        return sorted(
            [account for account in self.accounts if account.root == address],
            key=lambda account: int(
                re.compile(r"\D+").sub(
                    "", account.path if account.path is not None else ""
                )
            ),
        )

    def recursive_append_children(
        self, element: Optional[Union[Category, Account]], item: TreeItem
    ) -> None:
        """
        Add children items from element to item

        :param element: Account or Category instance or None
        :param item: TreeItem instance
        :return:
        """
        if element is None or isinstance(element, Category):
            parent_category_id = None if element is None else element.id

            # list of categories in category
            for category in self.category_children(parent_category_id):
                category_item = TreeItem(category, self.root_item)
                item.add_child(category_item)
                self.recursive_append_children(category, category_item)

            # list of accounts in category
            for account in self.account_list_by_category_id(parent_category_id):
                account_item = TreeItem(account, item)
                item.add_child(account_item)
                self.recursive_append_children(account, account_item)

        elif isinstance(element, Account):
            # list of derivation from this account
            for account in self.get_derivation_accounts(element.address):
                account_item = TreeItem(account, item)
                item.add_child(account_item)

    def index(
        self, row: int, column: int, parent: QModelIndex = QModelIndex()
    ) -> QModelIndex:
        """
        Return QModelIndex for row, column and parent

        :param row: Row index
        :param column: Column index
        :param parent: Parent QModelIndex instance
        :return:
        """
        if not parent or not parent.isValid():
            parent_item = self.root_item
        else:
            parent_item = parent.internalPointer()

        if not QAbstractItemModel.hasIndex(self, row, column, parent):
            return QModelIndex()

        child_item = parent_item.child(row)
        if child_item:
            return QAbstractItemModel.createIndex(self, row, column, child_item)

        return QModelIndex()

    def parent(self, child: QModelIndex) -> QModelIndex:  # type: ignore
        """
        Return parent QModelIndex of child QModelIndex

        :param child: QModelIndex instance
        :return:
        """
        if child.isValid():
            parent_item = child.internalPointer().parent()
            if parent_item:
                return QAbstractItemModel.createIndex(
                    self, parent_item.row(), 0, parent_item
                )
        return QModelIndex()

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        """
        Return row count of parent QModelIndex

        :param parent: QModelIndex instance
        :return:
        """
        if parent.isValid():
            return parent.internalPointer().child_count()
        return self.root_item.child_count()

    def columnCount(
        self, parent: QModelIndex = QModelIndex()  # pylint: disable=unused-argument
    ) -> int:
        """
        Return column count of parent QModelIndex

        :param parent: QModelIndex instance
        :return:
        """
        return self._column_count

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> QVariant:
        """
        Return data of cell for column index.column

        :param index: QModelIndex instance
        :param role: Item data role
        :return:
        """
        data = QVariant()
        if not index.isValid():
            return data

        tree_item = index.internalPointer()
        # display account address or category name in first column
        # only this column is indented
        if index.column() == 0 and role in (Qt.DisplayRole, Qt.EditRole):
            data = QVariant(tree_item.element.name)

        if (
            index.column() == 1
            and role == Qt.DisplayRole
            and isinstance(tree_item.element, Account)
        ):
            data = QVariant(tree_item.element.address)

        if role == Qt.FontRole and isinstance(tree_item.element, Account):
            if index.column() == 1:
                # display addresses in monospace font
                font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
                # if font is not found, choose equivalent monospace font on system
                font.setStyleHint(QFont.Monospace)
                # font.setBold(True)
            else:
                font = QFont()
            # display root account with underline
            if tree_item.element.root is None:
                font.setUnderline(True)
            return QVariant(font)

        # display wallet lock status icon for account in first column
        if (
            index.column() == 0
            and role == Qt.DecorationRole
            and isinstance(tree_item.element, Account)
        ):
            if self.application.wallets.exists(tree_item.element.address):
                data = (
                    QVariant(QIcon(QPixmap(ICON_ACCOUNT_WALLET_UNLOCKED)))
                    if self.application.wallets.is_unlocked(tree_item.element.address)
                    else QVariant(QIcon(QPixmap(ICON_ACCOUNT_WALLET_LOCKED)))
                )
            else:
                data = QVariant(QIcon(QPixmap(ICON_ACCOUNT_NO_WALLET)))

        # return icon of identity membership status
        if (
            index.column() == 0
            and role == MemberStatusRole
            and isinstance(tree_item.element, Account)
        ):
            if tree_item.element.identity_index is not None:
                identity = self.application.identities.get(
                    tree_item.element.identity_index
                )
                if identity is not None and identity.status == STATUS_VALIDATED:
                    data = QVariant(
                        QPixmap(ICON_IDENTITY).scaled(
                            16, 16, aspectRatioMode=Qt.KeepAspectRatio
                        )
                    )

        return data

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        """
        Return Qt.ItemFlags for index

        :param index: QModelIndex instance
        :return:
        """
        flags = super().flags(index)
        tree_item: TreeItem = index.internalPointer()

        if index.column() == 0:
            if isinstance(tree_item.element, Category):
                # category name is editable, can be dragged and dropped on
                return (
                    flags
                    | Qt.ItemIsEditable
                    | Qt.ItemIsDragEnabled
                    | Qt.ItemIsDropEnabled
                )
            # account can be dragged and dropped on
            return flags | Qt.ItemIsDragEnabled | Qt.ItemIsDropEnabled

        return flags

    def setData(self, index: QModelIndex, value: Any, role: int = 0) -> bool:
        """
        Set data after treeView editing

        :param index: QModelIndex instance
        :param value: Editor Value
        :param role: Role flag
        :return:
        """
        super().setData(index, role)

        tree_item: TreeItem = index.internalPointer()
        # if category name edited (not dropped where value is None)...
        if (
            role == Qt.EditRole
            and index.column() == 0
            and isinstance(tree_item.element, Category)
            and value is not None
        ):
            # update name
            tree_item.element.name = value
            # update repository
            self.application.categories.update(tree_item.element)

        self.dataChanged.emit(index, index)
        return True

    def insertRow(self, row: int, parent: QModelIndex = QModelIndex()) -> bool:
        """
        Insert row in parent index

        :param row: Index of row from 0 in parent children
        :param parent: QModelIndex instance
        :return:
        """
        self.beginInsertRows(parent, row, row)
        parent_item: TreeItem = parent.internalPointer()
        parent_id = None
        if parent_item is None:
            parent_item = self.root_item
        elif isinstance(parent_item.element, Category):
            parent_id = parent_item.element.id
        category = self.application.categories.create("", parent_id)
        success = parent_item.insert_child(row, TreeItem(category, parent_item))
        self.application.categories.repository.add(category)
        self.endInsertRows()
        return success

    def supportedDropActions(self):
        """
        Return supported drop action flags

        :return:
        """
        return Qt.MoveAction

    def mimeTypes(self) -> List[str]:
        """
        Return supported mime types for drag and drop

        :return:
        """
        return [self.mime_type]

    def mimeData(self, indexes: Iterable[QModelIndex]) -> QMimeData:
        """
        Return encoded mime-type values for indexes

        :param indexes: List of indexes
        :return:
        """
        sorted_indices = sorted(
            [index for index in indexes if index.isValid()],
            key=lambda index: index.row(),
        )

        data_list: List[Tuple[Union[Type[Category], Type[Account]], str]] = []
        for index in sorted_indices:
            if not index.isValid():
                continue
            if index.column() == 0:
                tree_item: TreeItem = index.internalPointer()
                if tree_item is None:
                    continue
                if isinstance(tree_item.element, Category):
                    data_list.append(
                        (tree_item.element.__class__, tree_item.element.id.hex)
                    )
                elif isinstance(tree_item.element, Account):
                    data_list.append(
                        (tree_item.element.__class__, tree_item.element.address)
                    )

        encoded_data = pickle.dumps(data_list)
        mime_data = QMimeData()
        mime_data.setData(self.mime_type, encoded_data)
        return mime_data

    def dropMimeData(
        self,
        data: QMimeData,
        action: Qt.DropAction,
        row: int,  # pylint: disable=unused-argument
        column: int,
        parent: QModelIndex,
    ) -> bool:
        """
        Handle dropped data

        :param data: QMimeData instance
        :param action: Action flag
        :param row: Row dropped on
        :param column: Column dropped on
        :param parent: QModelIndex instance
        :return:
        """
        if action == Qt.IgnoreAction:
            return True
        if not data.hasFormat(self.mime_type) or column > 0 or not parent.isValid():
            return False

        encoded_data = data.data(self.mime_type)
        decoded_data = pickle.loads(bytes(encoded_data))
        parent_tree_item: TreeItem = parent.internalPointer()

        # if dropped on no item or on item without element...
        if parent_tree_item is None or parent_tree_item.element is None:
            return False

        parent_category_id = None

        # if dropped on a category...
        if isinstance(parent_tree_item.element, Category):
            parent_category_id = parent_tree_item.element.id
        elif isinstance(parent_tree_item.element, Account):
            # if dropped on a derived account...
            if parent_tree_item.element.root is not None:
                # get root account category
                root_account = self.application.accounts.get_by_address(
                    parent_tree_item.element.root
                )
                if root_account is not None:
                    parent_category_id = root_account.category_id
                else:
                    return False
            else:
                # get root account category
                root_account = self.application.accounts.get_by_address(
                    parent_tree_item.element.address
                )
                if root_account is not None:
                    parent_category_id = root_account.category_id
                else:
                    return False

        for element_data in decoded_data:
            element_class, element_pk = element_data
            if element_class == Category:
                category = self.application.categories.get(UUID(hex=element_pk))
                if category is not None:
                    category.parent_id = parent_category_id
                    self.application.categories.repository.update(category)
            else:
                account = self.application.accounts.get_by_address(element_pk)
                if account is not None:
                    # if root account...
                    if account.root is None:
                        account.category_id = parent_category_id
                        self.application.accounts.repository.update(account)
                    else:
                        # get root account
                        root_account = self.application.accounts.get_by_address(
                            account.root
                        )
                        if root_account is not None:
                            root_account.category_id = parent_category_id
                            self.application.accounts.repository.update(root_account)
        return True



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List, Optional

from substrateinterface import Keypair

from tikka.domains.entities.smith import SmithCertification, SmithMembership
from tikka.domains.nodes import Nodes
from tikka.interfaces.adapters.network.smiths import NetworkSmithsInterface


class Smiths:

    """
    Smiths domain class
    """

    def __init__(
        self,
        nodes: Nodes,
        network: NetworkSmithsInterface,
    ):
        """
        Init Smiths domain

        :param nodes: Nodes domain instance
        :param network: NetworkSmithsInterface instance
        """
        self.nodes = nodes
        self.network = network

    def request_membership(self, keypair: Keypair, session_keys: str) -> bool:
        """
        Request a smith membership for the Keypair account with node session_keys

        :param keypair: Owner Keypair
        :param session_keys: Session public keys (hex string "0x123XYZ")
        :return:
        """
        return self.network.request_membership(keypair, session_keys)

    def revoke_membership(self, keypair: Keypair) -> bool:
        """
        Revoke a smith membership for the Keypair account

        :param keypair: Owner Keypair
        :return:
        """
        return self.network.revoke_membership(keypair)

    def claim_membership(self, keypair: Keypair) -> bool:
        """
        Claim that last smith membership request for the Keypair account fulfill all requirements

        :param keypair: Owner Keypair
        :return:
        """
        return self.network.claim_membership(keypair)

    def fetch_membership_from_network(
        self, identity_index: int
    ) -> Optional[SmithMembership]:
        """
        Fetch smith membership for identity index from network if any

        :param identity_index: Identity index
        :return:
        """
        return self.network.membership(identity_index)

    def fetch_pending_membership_from_network(self, identity_index: int) -> bool:
        """
        Fetch smith pending membership for identity index from network if any

        :param identity_index: Identity index
        :return:
        """
        return self.network.pending_membership(identity_index)

    def fetch_certs_by_receiver_from_network(
        self, receiver_address: str, receiver_identity_index: int
    ) -> Optional[List[SmithCertification]]:
        """
        Fetch certification (identity index, expire on block number) list for identity index from network if any

        :param receiver_address: Address of receiver account
        :param receiver_identity_index: Identity index of receiver
        :return:
        """
        return self.network.certs_by_receiver(receiver_address, receiver_identity_index)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from typing import Optional

import qrcode
from PyQt5.QtCore import QMutex, QPoint
from PyQt5.QtGui import QFont, QPixmap
from PyQt5.QtWidgets import QApplication, QMainWindow, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.address import DisplayAddress
from tikka.domains.entities.constants import AMOUNT_UNIT_KEY, DATA_PATH
from tikka.domains.entities.events import AccountEvent, TransferEvent, UnitEvent
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    ICON_ACCOUNT_NO_WALLET,
    ICON_ACCOUNT_WALLET_LOCKED,
    ICON_ACCOUNT_WALLET_UNLOCKED,
    SELECTED_UNIT_PREFERENCES_KEY,
)
from tikka.slots.pyqt.entities.qrcode_image import QRCodeImage
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.resources.gui.widgets.account_rc import Ui_AccountWidget
from tikka.slots.pyqt.widgets.account_menu import AccountPopupMenu
from tikka.slots.pyqt.windows.transfer import TransferWindow


class AccountWidget(QWidget, Ui_AccountWidget):
    """
    AccountWidget class
    """

    def __init__(
        self,
        application: Application,
        account: Account,
        mutex: QMutex,
        parent: Optional[QWidget] = None,
    ) -> None:
        """
        Init AccountWidget instance

        :param application: Application instance
        :param account: Account instance
        :param mutex: QMutex instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        self.account = account
        self.mutex = mutex

        self.monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        self.monospace_font.setStyleHint(QFont.Monospace)
        self.addressValueLabel.setFont(self.monospace_font)
        self.addressValueLabel.setText(self.account.address)

        # creating a pix map of qr code
        qr_code_pixmap = qrcode.make(
            self.account.address, image_factory=QRCodeImage
        ).pixmap()

        # set qrcode to the label
        self.QRCodeAddressLabel.setPixmap(qr_code_pixmap)

        self._update_ui()

        ##############################
        # ASYNC METHODS
        ##############################
        # Create a QWorker object
        self.fetch_from_network_async_qworker = AsyncQWorker(
            self.fetch_from_network, self.mutex
        )
        self.fetch_from_network_async_qworker.finished.connect(
            self._on_finished_fetch_from_network
        )

        # events
        self.refreshButton.clicked.connect(self.fetch_from_network_async_qworker.start)
        self.transferToButton.clicked.connect(self.transfer)
        self.customContextMenuRequested.connect(self.on_context_menu)

        # application events
        self.application.event_dispatcher.add_event_listener(
            AccountEvent.EVENT_TYPE_UPDATE, lambda e: self._update_ui()
        )
        self.application.event_dispatcher.add_event_listener(
            UnitEvent.EVENT_TYPE_CHANGED, lambda e: self._update_ui()
        )
        self.application.event_dispatcher.add_event_listener(
            TransferEvent.EVENT_TYPE_SENT, lambda e: self._on_transfer_sent()
        )

        self.fetch_from_network_async_qworker.start()

    def transfer(self):
        """
        When user click on transfer to button

        :return:
        """
        TransferWindow(
            self.application, self.mutex, None, self.account, parent=self.parent()
        ).exec_()

    def _on_transfer_sent(self):
        """
        Triggered after a successful transfer event

        :return:
        """
        # update account balance from network
        self.fetch_from_network_async_qworker.start()

    def fetch_from_network(self):
        """
        Fetch last account data from the network

        :return:
        """
        self.refreshButton.setEnabled(False)

        try:
            balance = self.application.accounts.fetch_balance_from_network(
                self.account.address
            )
        except Exception:
            # keep balance from cache DB if network failure
            balance = self.account.balance

        if balance is None:
            self.account.balance = 0
        else:
            self.account.balance = balance

        self.account.identity_index = (
            self.application.identities.fetch_index_from_network(self.account.address)
        )
        if self.account.identity_index is not None:
            self.application.identities.fetch_from_network(self.account.identity_index)

    def _on_finished_fetch_from_network(self):
        """
        Triggered when async request fetch_from_network is finished

        :return:
        """
        self.application.accounts.update(self.account)
        self.refreshButton.setEnabled(True)

    def _update_ui(self):
        """
        Update UI from self.account

        :return:
        """
        display_identity_status = {
            "Created": self._("Created"),
            "Validated": self._("Validated"),
            "ConfirmedByOwner": self._("Confirmed"),
            "Unknown": self._("Unknown"),
        }

        unit_preference = self.application.preferences_repository.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            amount = self.application.amounts.get_amount(unit_preference)
        else:
            amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)

        if self.account.name is None:
            self.nameLabel.setText("")
        else:
            self.nameLabel.setText(self.account.name)

        if self.account.balance is None:
            self.balanceLabel.setText("?")
        else:
            self.balanceLabel.setText(
                self.locale().toCurrencyString(
                    amount.value(self.account.balance), amount.symbol()
                )
            )

        if self.application.wallets.exists(self.account.address):
            if self.application.wallets.is_unlocked(self.account.address):
                self.lockStatusIcon.setPixmap(QPixmap(ICON_ACCOUNT_WALLET_UNLOCKED))
            else:
                self.lockStatusIcon.setPixmap(QPixmap(ICON_ACCOUNT_WALLET_LOCKED))
        else:
            self.lockStatusIcon.setPixmap(QPixmap(ICON_ACCOUNT_NO_WALLET))

        if self.account.identity_index is not None:
            identity = self.application.identities.get(self.account.identity_index)
            if identity.status not in display_identity_status:
                self.identityValueLabel.setText(display_identity_status["Unknown"])
            else:
                self.identityValueLabel.setText(
                    display_identity_status[identity.status]
                )
        else:
            self.identityValueLabel.setText(self._("None"))

        if self.account.root is None and self.account.path is None:
            self.derivationValueLabel.setText(self._("Root"))
        else:
            root_account = self.application.accounts.get_by_address(self.account.root)
            if root_account is not None and root_account.name is not None:
                self.derivationValueLabel.setFont(QFont())
                self.derivationValueLabel.setText(root_account.name + self.account.path)
            else:
                self.derivationValueLabel.setFont(self.monospace_font)
                self.derivationValueLabel.setText(
                    DisplayAddress(self.account.root).shorten + self.account.path
                )

    def on_context_menu(self, position: QPoint):
        """
        When right button on account tab

        :return:
        """
        # display popup menu at click position
        menu = AccountPopupMenu(self.application, self.account, self.mutex, self)
        menu.exec_(self.mapToGlobal(position))


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)
    account_ = Account("5GrwvaEF5zXb26Fz9rcQpDWS57CtERHpNehXCPcNoHGKutQY")

    main_window = QMainWindow()
    main_window.show()

    main_window.setCentralWidget(
        AccountWidget(application_, account_, QMutex(), main_window)
    )

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from typing import List, Optional

from substrateinterface import Keypair
from substrateinterface.exceptions import SubstrateRequestException

from tikka.domains.entities.smith import SmithCertification, SmithMembership
from tikka.interfaces.adapters.network.smiths import NetworkSmithsInterface


class NetworkSmiths(NetworkSmithsInterface):
    """
    NetworkSmiths class
    """

    def request_membership(self, keypair: Keypair, session_keys: str) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkSmithsInterface.request_membership.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            return False

        try:
            call = self.connections.rpc.client.compose_call(
                call_module="SmithMembership",
                call_function="request_membership",
                call_params=None,
            )
        except Exception as exception:
            logging.exception(exception)
            return False

        try:
            extrinsic = self.connections.rpc.client.create_signed_extrinsic(
                call=call, keypair=keypair
            )
        except Exception as exception:
            logging.exception(exception)
            return False

        try:
            # fixme: code stuck infinitely if no blocks are created on blockchain
            #       should have a timeout option
            result = self.connections.rpc.client.submit_extrinsic(
                extrinsic, wait_for_inclusion=True
            )
            logging.debug(
                "Extrinsic '%s' sent and included in block '%s'",
                result.extrinsic_hash,
                result.block_hash,
            )
        except SubstrateRequestException as exception:
            logging.exception(exception)
            return False

        if result.is_success is False:
            logging.error(result.error_message)

        return result.is_success

    def claim_membership(self, keypair: Keypair) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkSmithsInterface.claim_membership.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            return False

        try:
            call = self.connections.rpc.client.compose_call(
                call_module="SmithMembership",
                call_function="claim_membership",
                call_params=None,
            )
        except Exception as exception:
            logging.exception(exception)
            return False

        try:
            extrinsic = self.connections.rpc.client.create_signed_extrinsic(
                call=call, keypair=keypair
            )
        except Exception as exception:
            logging.exception(exception)
            return False

        try:
            # fixme: code stuck infinitely if no blocks are created on blockchain
            #       should have a timeout option
            result = self.connections.rpc.client.submit_extrinsic(
                extrinsic, wait_for_inclusion=True
            )
            logging.debug(
                "Extrinsic '%s' sent and included in block '%s'",
                result.extrinsic_hash,
                result.block_hash,
            )
        except SubstrateRequestException as exception:
            logging.exception(exception)
            return False

        if result.is_success is False:
            logging.error(result.error_message)

        return result.is_success

    def revoke_membership(self, keypair: Keypair) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkSmithsInterface.revoke_membership.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            return False

        try:
            call = self.connections.rpc.client.compose_call(
                call_module="SmithMembership",
                call_function="revoke_membership",
                call_params=None,
            )
        except Exception as exception:
            logging.exception(exception)
            return False

        try:
            extrinsic = self.connections.rpc.client.create_signed_extrinsic(
                call=call, keypair=keypair
            )
        except Exception as exception:
            logging.exception(exception)
            return False

        try:
            # fixme: code stuck infinitely if no blocks are created on blockchain
            #       should have a timeout option
            result = self.connections.rpc.client.submit_extrinsic(
                extrinsic, wait_for_inclusion=True
            )
            logging.debug(
                "Extrinsic '%s' sent and included in block '%s'",
                result.extrinsic_hash,
                result.block_hash,
            )
        except SubstrateRequestException as exception:
            logging.exception(exception)
            return False

        if result.is_success is False:
            logging.error(result.error_message)

        return result.is_success

    def membership(self, identity_index: int) -> Optional[SmithMembership]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkSmithsInterface.membership.__doc__
        )
        if not self.connections.is_connected():
            return None
        if self.connections.rpc.client is None:
            return None
        try:
            result = self.connections.rpc.client.query(
                "SmithMembership", "Membership", [identity_index]
            )
        except Exception as exception:
            logging.exception(exception)
            return None

        membership = None
        if result.value is not None:
            membership = SmithMembership(expire_on=result.value["expire_on"])

        return membership

    def pending_membership(self, identity_index: int) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkSmithsInterface.pending_membership.__doc__
        )
        if not self.connections.is_connected():
            return False
        if self.connections.rpc.client is None:
            return False

        try:
            result = self.connections.rpc.client.query(
                "SmithMembership", "PendingMembership", [identity_index]
            )
        except Exception as exception:
            logging.exception(exception)
            return False

        if result == ():
            return True

        return False

    def certs_by_receiver(
        self, receiver_address: str, receiver_identity_index: int
    ) -> Optional[List[SmithCertification]]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkSmithsInterface.certs_by_receiver.__doc__
        )
        if not self.connections.is_connected():
            return None
        if self.connections.rpc.client is None:
            return None
        try:
            result = self.connections.rpc.client.query(
                "SmithCert", "CertsByReceiver", [receiver_identity_index]
            )
        except Exception as exception:
            logging.exception(exception)
            return None

        storage_keys = []
        for issuer_identity_index, cert_expire_on_block in result.value:
            storage_keys.append(
                self.connections.rpc.client.create_storage_key(
                    "Identity", "Identities", [issuer_identity_index]
                )
            )

        multi_result = self.connections.rpc.client.query_multi(storage_keys)

        certifications = []
        for index, (storage_key, value_obj) in enumerate(multi_result):
            certifications.append(
                SmithCertification(
                    issuer_identity_index=storage_keys[index].params[0],
                    issuer_address=value_obj["owner_key"],
                    receiver_identity_index=receiver_identity_index,
                    receiver_address=receiver_address,
                    expire_on_block=result.value[index][1],
                )
            )

        return certifications



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/widgets/nodes.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_NodesWidget(object):
    def setupUi(self, NodesWidget):
        NodesWidget.setObjectName("NodesWidget")
        NodesWidget.resize(724, 526)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(NodesWidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.addServerButton = QtWidgets.QPushButton(NodesWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.addServerButton.sizePolicy().hasHeightForWidth())
        self.addServerButton.setSizePolicy(sizePolicy)
        self.addServerButton.setObjectName("addServerButton")
        self.verticalLayout_2.addWidget(self.addServerButton, 0, QtCore.Qt.AlignRight)
        self.nodesTableView = QtWidgets.QTableView(NodesWidget)
        self.nodesTableView.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.nodesTableView.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.nodesTableView.setShowGrid(True)
        self.nodesTableView.setSortingEnabled(False)
        self.nodesTableView.setObjectName("nodesTableView")
        self.nodesTableView.horizontalHeader().setStretchLastSection(True)
        self.nodesTableView.verticalHeader().setVisible(False)
        self.verticalLayout_2.addWidget(self.nodesTableView)

        self.retranslateUi(NodesWidget)
        QtCore.QMetaObject.connectSlotsByName(NodesWidget)

    def retranslateUi(self, NodesWidget):
        #  _translate = QtCore.QCoreApplication.translate
        NodesWidget.setWindowTitle(_("Network nodes"))
        self.addServerButton.setText(_("Add server"))
import tikka.slots.pyqt.resources.icons.index_rc
